import geopandas as gpd
import pandas as pd
import pyproj
import shapely
from pydantic import BaseModel, Field, InstanceOf, field_validator

from ..utils.basic_service_types import BASIC_SERVICE_TYPES
from .service_type import ServiceType


class Service(BaseModel):
    service_type: ServiceType
    geometry: InstanceOf[shapely.Point]
    capacity: int

    def to_dict(self):
        return {"service_type": self.service_type.name, "geometry": self.geometry, "capacity": self.capacity}

    @classmethod
    def from_gdf(cls, service_type: ServiceType, gdf) -> list:
        return [cls(service_type=service_type, **gdf.loc[i].to_dict()) for i in gdf.index]


class Town(BaseModel):
    id: int
    name: str
    population: int = Field(gt=0)
    geometry: InstanceOf[shapely.Point]
    services: list[Service] = []

    def __contains__(self, service_type: ServiceType) -> bool:
        """Returns True if service type is contained in town"""
        return service_type in self._capacities

    def __getitem__(self, service_type: ServiceType) -> dict[str, int]:
        """Get service type capacity and demand of the town"""
        result = {"capacity": 0, "demand": service_type.calculate_in_need(self.population)}
        if service_type in self._capacities:
            result["capacity"] = self._capacities[service_type]
        return result

    def to_dict(self):
        res = {"id": self.id, "name": self.name, "population": self.population, "geometry": self.geometry}
        df = pd.DataFrame([service.to_dict() for service in self.services])
        if not df.empty:
            for service_type, group in df.groupby("service_type"):
                res.update(
                    {
                        f"capacity_{service_type}": group["capacity"].sum(),
                        # f'count_{service_type}': len(group)
                    }
                )
        return res

    @classmethod
    def from_gdf(cls, gdf) -> dict:
        return {i: cls(id=i, **gdf.loc[i].to_dict()) for i in gdf.index}

    def update_services(self, service_type: ServiceType, gdf: gpd.GeoDataFrame | None = None):
        services = filter(lambda s: s.service_type != service_type, self.services)
        if gdf is not None:
            new_services = Service.from_gdf(service_type, gdf)
            services = [*services, *new_services]
        self.services = list(services)



import geopandas as gpd
import osmnx as ox
from pydantic import BaseModel, InstanceOf, field_validator

from ..models.service_type import ServiceType


KMH_TO_MM = 1_000 / 60


class Fetcher(BaseModel):

    units: InstanceOf[gpd.GeoDataFrame]

    @field_validator("units", mode="after")
    @staticmethod
    def validate_units(gdf):
        assert gdf.geom_type.isin(["Polygon", "MultiPolygon"]).all(), "geometry should be Polygon or MultiPolygon"
        return gdf[["geometry"]]

    @property
    def crs(self):
        return self.units.crs

    @property
    def geometry(self):
        return self.units.to_crs(4326).unary_union

    def fetch_towns(self):
        gdf = ox.features_from_polygon(self.geometry, tags={"place": ["town", "city", "village", "hamlet"]})
        gdf = gdf.to_crs(self.crs)
        gdf.geometry = gdf.representative_point()
        gdf["population"] = gdf["population"].fillna(0)
        gdf["population"] = gdf["population"].apply(lambda s: str(s).split(" ", maxsplit=1)[0]).apply(int)
        gdf = gdf[gdf["population"] > 0].reset_index()
        return gdf[["geometry", "name", "population"]]

    def fetch_services(self, service_type: ServiceType | str, capacity: int = 250) -> gpd.GeoDataFrame:
        gdf = ox.features_from_polygon(self.geometry, tags=service_type.osm_tags)
        gdf = gdf.reset_index().to_crs(self.crs)
        gdf["capacity"] = capacity
        gdf.geometry = gdf.representative_point()
        return gdf[["geometry", "capacity"]]

    def fetch_graph(self, speed=60):
        speed *= KMH_TO_MM
        graph = ox.graph_from_polygon(self.geometry, network_type="drive", simplify=True)
        for _, _, data in graph.edges(data=True):
            data["time_min"] = data["length"] / speed
        return graph



"""
Provision assessment for cities of an urban region
"""

import importlib


__version__ = importlib.metadata.version("townsnet")
__author__ = ""
__email__ = ""
__credits__ = []
__license__ = "BSD-3"

from .method import *
from .models import *
from .preprocessing import *
from .utils import *



BASIC_PLUS_SERVICE_TYPES = {
    "education": [
        {
            "name": "community_center",
            "name_ru": "центр детского творчества",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 65,
            "osm_tags": {"amenity": "community_centre"},
        },
    ],
    "healthcare": [
        {
            "name": "polyclinic",
            "name_ru": "поликлиника / центр семейной медицины",
            "weight": 0.3,
            "accessibility": 10,
            "demand": 13,
            "osm_tags": {"amenity": "clinic"},
        },
        {
            "name": "local_hospital",
            "name_ru": "участковая больница",
            "weight": 0.3,
            "accessibility": 60,
            "demand": 9,
            "osm_tags": {"amenity": "hospital"},
        },
    ],
    "commerce": [
        {
            "name": "supermarket",
            "name_ru": "ТЦ / супермаркет",
            "weight": 0.5,
            "accessibility": 15,
            "demand": 900,
            "osm_tags": {"shop": "supermarket"},
        },
        {
            "name": "zoo",
            "name_ru": "зоомагазин",
            "weight": 0.1,
            "accessibility": 60,
            "demand": 2,
            "osm_tags": {"shop": "pet"},
        },
    ],
    "catering": [
        {
            "name": "bar_restaurant",
            "name_ru": "бар / ресторан",
            "weight": 0.3,
            "accessibility": 30,
            "demand": 100,
            "osm_tags": {"amenity": ["bar", "restaurant"]},
        },
    ],
    "leisure": [
        {
            "name": "community_centre_culture_house",
            "name_ru": "комьюнити-центр / дом культуры",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 100,
            "osm_tags": {"amenity": "community_centre"},
        },
    ],
    "recreation": [
        {
            "name": "public_space",
            "name_ru": "общественные пространства",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 100,
            "osm_tags": {"place": "square"},
        },
        {
            "name": "park",
            "name_ru": "парк",
            "weight": 0.4,
            "accessibility": 30,
            "demand": 150,
            "osm_tags": {"leisure": "park"},
        },
    ],
    "sport": [
        {
            "name": "gym__fitness_center",
            "name_ru": "спортзал ОП / фитнес-центр",
            "weight": 0.2,
            "accessibility": 45,
            "demand": 25,
            "osm_tags": {"leisure": ["sports_hall", "sports_centre", "fitness_centre"]},
        },
        {
            "name": "skatepark__workout_for_teenagers",
            "name_ru": "скейтпарк / воркаут для подростков",
            "weight": 0.1,
            "accessibility": 45,
            "demand": 25,
            "osm_tags": {"sport": "skateboard"},
        },
    ],
    "service": [
        {
            "name": "domestic_services",
            "name_ru": "бытовые услуги",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 75,
            "osm_tags": {"amenity": "*"},
        },
        {
            "name": "bank",
            "name_ru": "отделение банка",
            "weight": 0.2,
            "accessibility": 30,
            "demand": 20,
            "osm_tags": {"amenity": "bank"},
        },
    ],
    "transport": [
        {
            "name": "gas_station",
            "name_ru": "автозаправка",
            "weight": 0.3,
            "accessibility": 60,
            "demand": 33,
            "osm_tags": {"amenity": "bank"},
        },
    ],
    "safeness": [
        {
            "name": "police_supporting_point",
            "name_ru": "опорный пункт полиции",
            "weight": 0.5,
            "accessibility": 60,
            "demand": 20,
            "osm_tags": {"amenity": "bank"},
        },
    ],
}



import pytest
from shapely import Polygon, Point, distance
import geopandas as gpd
import pandas as pd
from townsnet import Region, Provision

crs = 32636

grid_district_size = 50
grid_settlement_size = 10
grid_min = 0
grid_max = 100

default_capacity = 100

@pytest.fixture
def districts():
    polygons = []
    for x in range(grid_min, grid_max, grid_district_size):
        for y in range(grid_min, grid_max, grid_district_size):
            polygon = Polygon.from_bounds(x,y,x+grid_district_size, y+grid_district_size)
            polygons.append(polygon)
    gdf = gpd.GeoDataFrame(geometry=polygons, crs=crs)
    gdf['name'] = gdf.apply(lambda s : f"district_{s.name}", axis=1)
    return gdf

@pytest.fixture
def settlements():
    polygons = []
    for x in range(grid_min, grid_max, grid_settlement_size):
        for y in range(grid_min, grid_max, grid_settlement_size):
            polygon = Polygon.from_bounds(x,y,x+grid_settlement_size, y+grid_settlement_size)
            polygons.append(polygon)
    gdf = gpd.GeoDataFrame(geometry=polygons, crs=crs)
    gdf['name'] = gdf.apply(lambda s : f"settlement_{s.name}", axis=1)
    return gdf

@pytest.fixture
def towns(settlements):
    gdf = settlements.copy()
    gdf.geometry = gdf.geometry.representative_point()
    gdf['name'] = gdf.apply(lambda s : f"town_{s.name}", axis=1)
    gdf['population'] = gdf.apply(lambda s : (s.name+1)*1000, axis=1)
    return gdf

@pytest.fixture
def territories():
    polygons = []
    for x in range(grid_min, grid_max, grid_district_size):
        for y in range(grid_min, grid_max, grid_district_size):
            polygon = Polygon.from_bounds(x+grid_settlement_size/4,y+grid_settlement_size/4,x+grid_settlement_size/2, y+grid_settlement_size/2)
            polygons.append(polygon)
    gdf = gpd.GeoDataFrame(geometry=polygons, crs=crs)
    gdf['name'] = gdf.apply(lambda s : f"territory_{s.name}", axis=1)
    return gdf

@pytest.fixture
def accessibility_matrix(towns):
    acc_mx = pd.DataFrame(0.0, index=towns.index, columns=towns.index)
    for i in acc_mx.index:
        for j in acc_mx.columns:
            acc_mx.loc[i,j] = distance(towns.loc[i, 'geometry'], towns.loc[j, 'geometry'])/4
    return acc_mx

@pytest.fixture
def region(districts, settlements, towns, accessibility_matrix, territories):
    region = Region(districts=districts, settlements=settlements, towns=towns, accessibility_matrix=accessibility_matrix, territories=territories)
    for service_type in region.service_types:
        gdf = region.get_towns_gdf()[['geometry']]
        gdf['capacity'] = default_capacity
        gdf = region.match_services_towns(gdf)
        region.update_services(service_type, gdf)
    return region

@pytest.fixture
def provisions(region):
    prov = Provision(region=region)
    provs = {}
    for st in region.service_types:
        provs[st] = prov.calculate(st)
    return provs

def test_provisions(provisions, region):
    for prov in provisions.values():
        districts_gdf, settlements_gdf, towns_gdf, links_gdf = prov
        assert all(districts_gdf.index == region.districts.index)
        assert all(settlements_gdf.index == region.settlements.index)
        assert all(towns_gdf.index == region.get_towns_gdf().index)
        for gdf in [districts_gdf, settlements_gdf, towns_gdf]:
            assert all(gdf.provision>=0) and all(gdf.provision<= 1)
            assert all(gdf.demand >= 0)
            assert all(gdf.demand_left >= 0)
            assert all(gdf.demand_within >= 0)
            assert all(gdf.demand_without >= 0)
            assert all(gdf.capacity >= 0)
            assert all(gdf.capacity_left >= 0)
        assert all(links_gdf.demand >= 0)
        assert all(links_gdf['from'].isin(region.get_towns_gdf().index))
        assert all(links_gdf['to'].isin(region.get_towns_gdf().index))

def test_context_provision(region, provisions):
    for territory in region.territories:
        _, value = territory.get_indicators(provisions)
        assert 0 <= value <= 11


import geopandas as gpd
import pandas as pd
import shapely
from pydantic import BaseModel, InstanceOf

from .service_type import ServiceType


class Territory(BaseModel):
    id: int
    name: str
    geometry: InstanceOf[shapely.Polygon]

    @staticmethod
    def _aggregate_provision(prov_gdf: gpd.GeoDataFrame):
        columns = ["demand", "demand_left", "demand_within", "demand_without", "capacity", "capacity_left"]
        result = {column: prov_gdf[column].sum() for column in columns}
        result["provision"] = result["demand_within"] / result["demand"]
        return result

    def get_context_provision(
        self, service_type: ServiceType, towns_gdf: gpd.GeoDataFrame, speed: float = 283.33
    ) -> tuple[gpd.GeoDataFrame, dict[str, float]]:

        buffer_meters = service_type.accessibility * speed

        towns_gdf = towns_gdf.copy()
        towns_gdf["distance"] = towns_gdf["geometry"].apply(lambda g: shapely.distance(g, self.geometry))
        towns_gdf = towns_gdf[towns_gdf["distance"] <= buffer_meters]
        towns_gdf = towns_gdf.rename(columns={"index_right": "town_id"})

        territory_dict = self._aggregate_provision(towns_gdf)
        territory_dict["buffer_meters"] = buffer_meters
        return towns_gdf, territory_dict

    def get_indicators(self, provs: dict[ServiceType, tuple], min_provision=0.8):
        dicts = []
        for service_type, prov_tuple in provs.items():
            _, _, towns_gdf, _ = prov_tuple
            _, territory_dict = self.get_context_provision(service_type, towns_gdf)
            dicts.append(
                {
                    "category": service_type.category.name,
                    "infrastructure": service_type.infrastructure.name,
                    "service_type": service_type.name,
                    "weight": service_type.weight,
                    **territory_dict,
                }
            )
        indicators = pd.DataFrame(dicts).set_index("service_type", drop=True)
        indicators["assessment"] = indicators.apply(
            lambda s: s["weight"] if s["provision"] >= min_provision else 0, axis=1
        )
        basic, basic_plus, comfort = indicators.groupby("category").agg({"assessment": "sum"})["assessment"]
        return indicators, basic + basic_plus / 4 + comfort / 4

    def to_dict(self):
        return {"id": self.id, "name": self.name, "geometry": self.geometry}

    @classmethod
    def from_gdf(cls, gdf: gpd.GeoDataFrame) -> list:
        return {i: cls(id=i, **gdf.loc[i].to_dict()) for i in gdf.index}



from .basic_plus_service_types import BASIC_PLUS_SERVICE_TYPES
from .basic_service_types import BASIC_SERVICE_TYPES
from .comfort_service_types import COMFORT_SERVICE_TYPES


SERVICE_TYPES = [
    *[
        {"category": "basic", "infrastructure": infrastructure, **st}
        for infrastructure, stypes in BASIC_SERVICE_TYPES.items()
        for st in stypes
    ],
    *[
        {"category": "basic_plus", "infrastructure": infrastructure, **st}
        for infrastructure, stypes in BASIC_PLUS_SERVICE_TYPES.items()
        for st in stypes
    ],
    *[
        {"category": "comfort", "infrastructure": infrastructure, **st}
        for infrastructure, stypes in COMFORT_SERVICE_TYPES.items()
        for st in stypes
    ],
]
